@@ -4,4 +4,6 @@ DEFAULT_SERVER_CAPABILITIES = (
     Capabilities.CLIENT_PROTOCOL_41
     | Capabilities.CLIENT_DEPRECATE_EOF
     | Capabilities.CLIENT_CONNECT_WITH_DB
+    | Capabilities.CLIENT_QUERY_ATTRIBUTES
+    | Capabilities.CLIENT_CONNECT_ATTRS
 )
